@@ -1,8 +1,7 @@
 from abc import abstractmethod
 from functools import partial, wraps
-from typing import Any, Callable, Union
+from typing import Any, Callable, Sequence, Type, TypeVar
 
-import cachetools
 import numpy as np
 import pandas as pd
 import scipy.special
@@ -10,9 +9,13 @@ from scipy.optimize import minimize
 from scipy.stats import beta, expon, gamma, genextreme, genpareto, norm, pareto, uniform
 
 from pykelihood import kernels
+from pykelihood.generic_types import Obs
+from pykelihood.metrics import opposite_log_likelihood
 from pykelihood.parameters import ConstantParameter, Parametrized
-from pykelihood.stats_utils import ConditioningMethod
-from pykelihood.utils import hash_with_series, ifnone
+from pykelihood.utils import ifnone
+
+T = TypeVar("T")
+SomeDistribution = TypeVar("SomeDistribution", bound="Distribution")
 
 EULER = -scipy.special.psi(1)
 
@@ -22,66 +25,47 @@ class Distribution(Parametrized):
         return (self.__class__.__name__,) + self.params
 
     @abstractmethod
-    def rvs(self, size: int):
+    def rvs(self, size: int, *args, **kwargs) -> np.ndarray:
         return NotImplemented
 
     @abstractmethod
-    def cdf(self, x: Union[np.array, float]):
+    def cdf(self, x: Obs):
         return NotImplemented
 
     @abstractmethod
-    def isf(self, q: float):
+    def isf(self, q: Obs):
         return NotImplemented
 
     @abstractmethod
-    def ppf(self, q: float):
+    def ppf(self, q: Obs):
         return NotImplemented
 
     @abstractmethod
-    def pdf(self, x: Union[np.array, float]):
+    def pdf(self, x: Obs):
         return NotImplemented
 
     @classmethod
     def param_dict_to_vec(cls, x: dict):
         return tuple(x.get(p) for p in cls.params_names)
 
-    def sf(self, x: Union[np.array, float]):
+    def sf(self, x: Obs):
         return 1 - self.cdf(x)
 
-    def logcdf(self, x: Union[np.array, float]):
+    def logcdf(self, x: Obs):
         return np.log(self.cdf(x))
 
-    def logsf(self, x: Union[np.array, float]):
+    def logsf(self, x: Obs):
         return np.log(self.sf(x))
 
-    def logpdf(self, x: Union[pd.Series, np.array, float], *args, **kwds):
-        return np.log(self.pdf(x, *args, **kwds))
+    def logpdf(self, x: Obs):
+        return np.log(self.pdf(x))
 
-    def inverse_cdf(self, q: float):
+    def inverse_cdf(self, q: Obs):
         if hasattr(self, "ppf"):
             return self.ppf(q)
         else:
             return self.isf(1 - q)
 
-    def log_likelihood(
-        self,
-        data: Union[np.array, pd.Series],
-        penalty: Callable = ConditioningMethod.no_conditioning,
-        *args,
-        **kwds,
-    ):
-        res = self.logpdf(data, *args, **kwds)
-        return np.sum(res) - penalty(data, self)
-
-    def opposite_log_likelihood(
-        self,
-        data: Union[np.array, pd.Series],
-        penalty: Callable = ConditioningMethod.no_conditioning,
-        *args,
-        **kwds,
-    ):
-        return -self.log_likelihood(data, penalty, *args, **kwds)
-
     def _process_fit_params(self, **kwds):
         param_dict = self.param_dict.copy()
         for name in param_dict:
@@ -95,32 +79,35 @@ class Distribution(Parametrized):
                     param_dict[name] = param_dict[name].with_params(subdic.values())
         return param_dict
 
+    def _apply_constraints(self, data):
+        return data
+
     @classmethod
     def fit(
-        cls,
-        data,
-        x0=None,
-        penalty=ConditioningMethod.no_conditioning,
+        cls: Type[SomeDistribution],
+        data: Obs,
+        x0: Sequence[float] = None,
+        score: Callable[["Distribution", Obs], float] = opposite_log_likelihood,
         **fixed_values,
-    ):
+    ) -> SomeDistribution:
         init_parms = {}
         for k in cls.params_names:
             if k in fixed_values:
-                v = fixed_values[k]
+                v = fixed_values.pop(k)
                 if isinstance(v, Parametrized):
                     init_parms[k] = v
                 else:
                     init_parms[k] = ConstantParameter(v)
-        init = cls(**init_parms)
+        init = cls(**init_parms, **fixed_values)
         x0 = x0 if x0 is not None else init.optimisation_params
         if len(x0) != len(init.optimisation_params):
             raise ValueError(
                 f"Expected {len(init.optimisation_params)} values in x0, got {len(x0)}"
             )
+        data = init._apply_constraints(data)
 
-        def to_minimize(x):
-            o = init.with_params(x)
-            return o.opposite_log_likelihood(data, penalty=penalty)
+        def to_minimize(x) -> float:
+            return score(init.with_params(x), data)
 
         res = minimize(to_minimize, x0, method="Nelder-Mead")
         return init.with_params(res.x)
@@ -128,13 +115,13 @@ class Distribution(Parametrized):
     def fit_instance(
         self,
         data,
-        penalty: Callable = ConditioningMethod.no_conditioning,
         fixed_params=None,
+        score=opposite_log_likelihood,
         **kwds,
     ):
         param_dict = self._process_fit_params(**(fixed_params or {}))
         kwds.update(param_dict)
-        return self.fit(data, penalty=penalty, **kwds)
+        return self.fit(data, score=score, **kwds)
 
 
 class AvoidAbstractMixin(object):
@@ -172,15 +159,34 @@ class ScipyDistribution(Distribution, AvoidAbstractMixin):
 
         return g
 
+    def _wrapper(self, f, x, _is_rvs=True, **extra_args):
+        params = {}
+        other_args = {}
+        for key, value in extra_args.items():
+            if key in self.params_names:
+                params[key] = value
+            else:
+                other_args[key] = value
+        if not _is_rvs:
+            return f(x, **self._to_scipy_args(**params), **other_args)
+        return f(**self._to_scipy_args(**params), size=x, **other_args)
+
     def __getattr__(self, item):
-        if item not in ("pdf", "logpdf", "cdf", "logcdf", "ppf", "isf", "sf", "logsf"):
+        if item not in (
+            "pdf",
+            "logpdf",
+            "cdf",
+            "logcdf",
+            "ppf",
+            "isf",
+            "sf",
+            "logsf",
+            "rvs",
+        ):
             return super(ScipyDistribution, self).__getattr__(item)
         f = getattr(self.base_module, item)
-        g = partial(self._wrapper, f)
+        g = partial(self._wrapper, f, _is_rvs=(item == "rvs"))
         g = self._correct_trends(g)
-        g = cachetools.cached(
-            self._cache, key=lambda x: hash((item, hash_with_series(x)))
-        )(g)
         self.__dict__[item] = g
         return g
 
@@ -191,13 +197,9 @@ class Uniform(ScipyDistribution):
 
     def __init__(self, loc=0.0, scale=1.0):
         super(Uniform, self).__init__(loc, scale)
-        self._cache = {}
 
-    def _wrapper(self, f, x, loc=None, scale=None):
-        return f(x, ifnone(loc, self.loc()), ifnone(scale, self.scale()))
-
-    def rvs(self, size, loc=None, scale=None):
-        return uniform.rvs(ifnone(loc, self.loc()), ifnone(scale, self.scale()), size)
+    def _to_scipy_args(self, loc=None, scale=None):
+        return {"loc": ifnone(loc, self.loc()), "scale": ifnone(scale, self.scale())}
 
 
 class Exponential(ScipyDistribution):
@@ -206,17 +208,11 @@ class Exponential(ScipyDistribution):
 
     def __init__(self, loc=0.0, rate=1.0):
         super(Exponential, self).__init__(loc, rate)
-        self._cache = {}
 
-    def _wrapper(self, f, x, loc=None, rate=None):
+    def _to_scipy_args(self, loc=None, rate=None):
         if rate is not None:
             rate = 1 / rate
-        return f(x, ifnone(loc, self.loc()), ifnone(rate, 1 / self.rate()))
-
-    def rvs(self, size, loc=None, rate=None):
-        return self.base_module.rvs(
-            ifnone(loc, self.loc()), ifnone(rate, 1 / self.rate()), size
-        )
+        return {"loc": ifnone(loc, self.loc()), "scale": ifnone(rate, 1 / self.rate())}
 
 
 class Gamma(ScipyDistribution):
@@ -225,23 +221,13 @@ class Gamma(ScipyDistribution):
 
     def __init__(self, loc=0.0, scale=1.0, shape=0.0):
         super(Gamma, self).__init__(loc, scale, shape)
-        self._cache = {}
 
-    def _wrapper(self, f, x, loc=None, scale=None, shape=None):
-        return f(
-            x,
-            ifnone(shape, self.shape()),
-            ifnone(loc, self.loc()),
-            ifnone(scale, self.scale()),
-        )
-
-    def rvs(self, size, loc=None, scale=None, shape=None):
-        return self.base_module.rvs(
-            ifnone(shape, self.shape()),
-            ifnone(loc, self.loc()),
-            ifnone(scale, self.scale()),
-            size,
-        )
+    def _to_scipy_args(self, loc=None, scale=None, shape=None):
+        return {
+            "a": ifnone(shape, self.shape()),
+            "loc": ifnone(loc, self.loc()),
+            "scale": ifnone(scale, self.scale()),
+        }
 
 
 class Pareto(ScipyDistribution):
@@ -250,23 +236,13 @@ class Pareto(ScipyDistribution):
 
     def __init__(self, loc=0.0, scale=1.0, alpha=1.0):
         super(Pareto, self).__init__(loc, scale, alpha)
-        self._cache = {}
 
-    def _wrapper(self, f, x, loc=None, scale=None, alpha=None):
-        return f(
-            x,
-            ifnone(alpha, self.alpha()),
-            ifnone(loc, self.loc()),
-            ifnone(scale, self.scale()),
-        )
-
-    def rvs(self, size, loc=None, scale=None, alpha=None):
-        return self.base_module.rvs(
-            ifnone(alpha, self.alpha()),
-            ifnone(loc, self.loc()),
-            ifnone(scale, self.scale()),
-            size,
-        )
+    def _to_scipy_args(self, loc=None, scale=None, alpha=None):
+        return {
+            "c": ifnone(alpha, self.alpha()),
+            "loc": ifnone(loc, self.loc()),
+            "scale": ifnone(scale, self.scale()),
+        }
 
 
 class Beta(ScipyDistribution):
@@ -275,25 +251,14 @@ class Beta(ScipyDistribution):
 
     def __init__(self, loc=0.0, scale=1.0, alpha=2.0, beta=1.0):
         super(Beta, self).__init__(loc, scale, alpha, beta)
-        self._cache = {}
 
-    def _wrapper(self, f, x, loc=None, scale=None, alpha=None, beta=None):
-        return f(
-            x,
-            ifnone(alpha, self.alpha()),
-            ifnone(beta, self.beta()),
-            ifnone(loc, self.loc()),
-            ifnone(scale, self.scale()),
-        )
-
-    def rvs(self, size, loc=None, scale=None, alpha=None, beta=None):
-        return self.base_module.rvs(
-            ifnone(alpha, self.alpha()),
-            ifnone(beta, self.beta()),
-            ifnone(loc, self.loc()),
-            ifnone(scale, self.scale()),
-            size,
-        )
+    def _to_scipy_args(self, loc=None, scale=None, alpha=None, beta=None):
+        return {
+            "a": ifnone(alpha, self.alpha()),
+            "b": ifnone(beta, self.beta()),
+            "loc": ifnone(loc, self.loc()),
+            "scale": ifnone(scale, self.scale()),
+        }
 
 
 class Normal(ScipyDistribution):
@@ -302,15 +267,9 @@ class Normal(ScipyDistribution):
 
     def __init__(self, loc=0.0, scale=1.0):
         super(Normal, self).__init__(loc, scale)
-        self._cache = {}
 
-    def _wrapper(self, f, x, loc=None, scale=None):
-        return f(x, ifnone(loc, self.loc()), ifnone(scale, self.scale()))
-
-    def rvs(self, size, loc=None, scale=None):
-        return self.base_module.rvs(
-            ifnone(loc, self.loc()), ifnone(scale, self.scale()), size
-        )
+    def _to_scipy_args(self, loc=None, scale=None):
+        return {"loc": ifnone(loc, self.loc()), "scale": ifnone(scale, self.scale())}
 
 
 class GEV(ScipyDistribution):
@@ -319,7 +278,6 @@ class GEV(ScipyDistribution):
 
     def __init__(self, loc=0.0, scale=1.0, shape=0.0):
         super(GEV, self).__init__(loc, scale, shape)
-        self._cache = {}
 
     def lb_shape(self, data):
         x_min = data.min()
@@ -341,25 +299,14 @@ class GEV(ScipyDistribution):
         else:
             return self.scale / (x_max - self.loc())
 
-    def _wrapper(self, f, x, loc=None, scale=None, shape=None):
+    def _to_scipy_args(self, loc=None, scale=None, shape=None):
         if shape is not None:
             shape = -shape
-        return f(
-            x,
-            ifnone(shape, -self.shape()),
-            ifnone(loc, self.loc()),
-            ifnone(scale, self.scale()),
-        )
-
-    def rvs(self, size, shape=None, loc=None, scale=None):
-        if shape is not None:
-            shape = -shape
-        return self.base_module.rvs(
-            ifnone(shape, -self.shape()),
-            ifnone(loc, self.loc()),
-            ifnone(scale, self.scale()),
-            size,
-        )
+        return {
+            "c": ifnone(shape, -self.shape()),
+            "loc": ifnone(loc, self.loc()),
+            "scale": ifnone(scale, self.scale()),
+        }
 
 
 class MixtureExponentialModel(Distribution):
@@ -367,7 +314,6 @@ class MixtureExponentialModel(Distribution):
 
     def __init__(self, theta=0.99):
         super(MixtureExponentialModel, self).__init__(theta)
-        self._cache = {}
 
     def rvs(self, size):
         theta = self.theta()
@@ -399,23 +345,13 @@ class GPD(ScipyDistribution):
 
     def __init__(self, loc=0.0, scale=1.0, shape=0.0):
         super(GPD, self).__init__(loc, scale, shape)
-        self._cache = {}
 
-    def _wrapper(self, f, x, loc=None, scale=None, shape=None):
-        return f(
-            x,
-            ifnone(shape, self.shape()),
-            ifnone(loc, self.loc()),
-            ifnone(scale, self.scale()),
-        )
-
-    def rvs(self, size, loc=None, scale=None, shape=None):
-        return self.base_module.rvs(
-            ifnone(shape, self.shape()),
-            ifnone(loc, self.loc()),
-            ifnone(scale, self.scale()),
-            size,
-        )
+    def _to_scipy_args(self, loc=None, scale=None, shape=None):
+        return {
+            "c": ifnone(shape, self.shape()),
+            "loc": ifnone(loc, self.loc()),
+            "scale": ifnone(scale, self.scale()),
+        }
 
 
 class ExtendedGPD(Distribution):
@@ -431,20 +367,20 @@ class ExtendedGPD(Distribution):
     def rvs(self, size: int):
         return self.inverse_cdf(Uniform().rvs(size))
 
-    def cdf(self, x: Union[np.array, float]):
+    def cdf(self, x: Obs):
         return 1 - np.clip(
             1 + self.shape * ((x - self.loc) / self.scale), 0, a_max=None
         ) ** (-1 / self.shape)
 
-    def inverse_cdf(self, q: float):
+    def inverse_cdf(self, q: Obs):
         return self.loc + self.scale * ((1 - q) ** (-self.shape) - 1) / self.shape
 
-    def pdf(self, x: Union[np.array, float]):
+    def pdf(self, x: Obs):
         return (1 / self.scale) * np.clip(
             1 + self.shape * ((x - self.loc) / self.scale), 0, a_max=None
         ) ** (-1 / self.shape - 1)
 
-    def isf(self, q: float):
+    def isf(self, q: Obs):
         return self.loc + self.scale * (q ** (-self.shape) - 1) / self.shape
 
 
@@ -519,18 +455,18 @@ class PointProcess(Distribution):
     def fit_instance(
         self,
         data,
-        penalty: Callable = ConditioningMethod.no_conditioning,
         fixed_params=None,
+        score=opposite_log_likelihood,
         **kwds,
     ):
         kwds.update(**(fixed_params or {}))
-        return self.fit(data, penalty=penalty, **kwds)
+        return self.fit(data, score=score, **kwds)
 
     def fit(
         self,
         data,
-        penalty=ConditioningMethod.no_conditioning,
         x0=None,
+        score=opposite_log_likelihood,
         opt_method="Nelder-Mead",
         plot=False,
         *args,
@@ -891,18 +827,18 @@ class CompositionDistribution(Distribution):
     def fit_instance(
         self,
         data,
-        penalty: Callable = ConditioningMethod.no_conditioning,
         fixed_params=None,
+        score=opposite_log_likelihood,
         **kwds,
     ):
         kwds.update(**(fixed_params or {}))
-        return self.fit(data, penalty=penalty, **kwds)
+        return self.fit(data, score=score, **kwds)
 
     def fit(
         self,
         data: pd.Series,
-        penalty=ConditioningMethod.no_conditioning,
         x0=None,
+        score=opposite_log_likelihood,
         opt_method="Nelder-Mead",
         *args,
         **kwds,
@@ -912,7 +848,7 @@ class CompositionDistribution(Distribution):
         x0 = obj.optimisation_params if x0 is None else x0
 
         def to_minimize(var_params):
-            return obj.with_params(var_params).opposite_log_likelihood(data, penalty)
+            return score(obj.with_params(var_params), data)
 
         result = minimize(
             to_minimize,
@@ -925,55 +861,62 @@ class CompositionDistribution(Distribution):
 
 
 class TruncatedDistribution(Distribution):
-    def __init__(self, distribution: Distribution, upper_bound=None, lower_bound=None):
-        super(TruncatedDistribution, self).__init__()
-        if upper_bound is None and lower_bound is None:
-            raise ValueError(
-                "A lower bound OR upper bound should be provided for the truncated distribution."
-            )
-        if upper_bound is None:
-            upper_bound = np.inf
-        elif lower_bound is None:
-            lower_bound = -np.inf
-        self.distribution = distribution
+    params_names = ("distribution",)
+
+    def __init__(
+        self, distribution: Distribution, lower_bound=-np.inf, upper_bound=np.inf
+    ):
+        if upper_bound == lower_bound:
+            raise ValueError("Both bounds are equal.")
+        super(TruncatedDistribution, self).__init__(distribution)
         self.lower_bound = lower_bound
         self.upper_bound = upper_bound
+        lower_cdf = self.distribution.cdf(self.lower_bound)
+        upper_cdf = self.distribution.cdf(self.upper_bound)
+        self._normalizer = upper_cdf - lower_cdf
 
-    def rvs(self, size: int):
+    def _build_instance(self, **new_params):
+        distribution = new_params.pop("distribution")
+        if new_params:
+            raise ValueError(f"Unexpected arguments: {new_params}")
+        return type(self)(distribution, self.lower_bound, self.upper_bound)
+
+    def _valid_indices(self, x: np.ndarray):
+        return (self.lower_bound <= x) & (x <= self.upper_bound)
+
+    def _apply_constraints(self, x):
+        return x[self._valid_indices(x)]
+
+    def fit_instance(self, *args, **kwargs):
+        kwargs.update(lower_bound=self.lower_bound, upper_bound=self.upper_bound)
+        return super().fit_instance(*args, **kwargs)
+
+    def rvs(self, size: int, *args, **kwargs):
         u = Uniform(
             self.distribution.cdf(self.lower_bound),
             self.distribution.cdf(self.upper_bound),
         )
-        return self.distribution.inverse_cdf(u.rvs(size))
+        return self.distribution.inverse_cdf(u.rvs(size, *args, **kwargs))
 
-    def pdf(self, x: Union[np.array, float]):
+    def pdf(self, x: Obs):
         return np.where(
-            self.lower_bound <= x <= self.upper_bound,
-            self.distribution.pdf(x)
-            / (
-                self.distribution.cdf(self.upper_bound)
-                - self.distribution.cdf(self.lower_bound)
-            ),
+            self._valid_indices(x),
+            self.distribution.pdf(x) / self._normalizer,
             0.0,
         )
 
-    def cdf(self, x: Union[np.array, float]):
-        denom = self.distribution.cdf(self.upper_bound) - self.distribution.cdf(
-            self.lower_bound
-        )
+    def cdf(self, x: Obs):
         right_range_x = (
             self.distribution.cdf(x) - self.distribution.cdf(self.lower_bound)
-        ) / denom
-        return np.where(self.lower_bound <= x <= self.upper_bound, right_range_x, 0.0)
+        ) / self._normalizer
+        return np.where(self._valid_indices(x), right_range_x, 0.0)
 
-    def isf(self, q: float):
-        mult = self.distribution.cdf(self.upper_bound) - self.distribution.cdf(
-            self.lower_bound
+    def isf(self, q: Obs):
+        return self.distribution.isf(
+            self.distribution.isf(self.upper_bound) + q * self._normalizer
         )
-        return self.distribution.isf(self.distribution.isf(self.upper_bound) + q * mult)
 
-    def ppf(self, q: float):
-        mult = self.distribution.cdf(self.upper_bound) - self.distribution.cdf(
-            self.lower_bound
+    def ppf(self, q: Obs):
+        return self.distribution.ppf(
+            self.distribution.cdf(self.lower_bound) + q * self._normalizer
         )
-        return self.distribution.ppf(self.distribution.cdf(self.lower_bound) + q * mult)
